# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2017,2018
"""
Stream transformations using relational predicates.
"""

from streamsx.spl.op import Invoke, Map

import streamsx.standard._version
__version__ = streamsx.standard._version.__version__

class Aggregate(Map):
    """Aggregation against a window of a structured schema stream.

    The resuting stream (attribute ``stream``) will contain aggregations
    of the window defined by the methods invoked against the instance.

    The aggregation is invoked when the window is triggered.
    A grouped aggregation will result in `N` output tuples per aggregation
    where `N` is the number of groups in the window at the time of the
    trigger. A non-grouped window results in a single tuple per aggregation.
    A window punctuation mark is submitted after the tuples resulting
    from the aggregation.

    In all examples, an input schema of ``tuple<int32 id, timestamp ts, float64 reading>`` is assumed representing an input stream of sensor readings.

    Example of aggregating sensor readings to produce a stream containing
    the maximum, minimum and average reading over the last ten minutes
    updating every minute and grouped by sensor::

        from streamsx.standard.relational import Aggregate

        # Declare the window
        win = readings.last(datetime.timedelta(minutes=10)).trigger(datetime.timedelta(minutes=1))

        # Declare the output schema
        schema = 'tuple<int32 id, timestamp ts, float64 max_reading, float64 min_reading, float64 avg_reading>'


        # Invoke the aggregation
        agg = Aggregate.invoke(win, schema, group='id')

        # Declare the output attribute assignments.
        agg.min_reading = agg.min('reading')
        agg.max_reading = agg.max('reading')
        agg.avg_reading = agg.average('reading')

        # resulting stream is agg.stream
        agg.stream.print()

    When an output attribute is not assigned and it has a matching
    input attribute the value in the last tuple for the group is used.
    In the example above ``id`` will be set to the sensor identifier
    of the group and ``ts`` will be set to the timestamp of the most
    recent tuple for the group.

    For the output attribute assignment methods that return a grouped
    value, such as :py:meth:`count`, :py:meth:`max`, when the window is not
    grouped the value is against all tuples in the window.

    The aggregation is implemented using the ``spl.relational::Aggregate``
    SPL primitive operator from the SPL Standard toolkit.
    """
    @staticmethod
    def invoke(window, schema, group=None, name=None):
        """Invoke an aggregation against a window.

        Args:
            window(streamsx.topology.topology.Window): Window to aggregate against.
            schema(str,StreamSchema): Schema of output stream containing aggregations.
            group(str): Attribute name to group aggregations.
            name(str): Invocation name, defaults to a generated name.

        Returns:
             Aggregate: Aggregate invocation.
        """
        if not isinstance(window, streamsx.topology.topology.Window):
            raise TypeError(window)
        _op = Aggregate(window, schema, group, name=name)
        if window._config is not None:
           if 'partitioned' in window._config:
              if (window._config['partitioned']):
                 _op.params['partitionBy'] =  _op.attribute(window._config['partitionBy'])
        return _op
  
    def _output_func(self, name, attribute=None):
        _eofn = name + '('
        if attribute is not None:
            _eofn = _eofn + attribute
        _eofn = _eofn + ')'
        return self.output(self.expression(_eofn))
        
    def __init__(self, window, schema, group=None, name=None):
        topology = window.topology
        kind="spl.relational::Aggregate"
        params = dict()
        if group is not None:
            params['groupBy'] = group
        #if aggregateIncompleteWindows is not None:
        #    params['aggregateIncompleteWindows'] = aggregateIncompleteWindows
        #if aggregateEvictedPartitions is not None:
        #    params['aggregateEvictedPartitions'] = aggregateEvictedPartitions
        super(Aggregate, self).__init__(kind,window,schema,params,name)

    def count(self):
        """Count of tuples in the group.

        Returns an output expression of type ``int32`` representing
        the number of tuples in the group.

        Example::

            # Count the number of tuples grouped by sensor id in the last minute
            schema = 'tuple<int32 id, timestamp ts, int32 n>'
            agg = Aggregate.invoke(s.last(datetime.timedelta(minutes=1)), schema, group='id')
            agg.n = agg.count()

        Returns:
            Expression: Output expression with the type ``int32``.
        """
        return self._output_func('Count')

    def count_all(self):
        """Count of all tuples in the window.

        Returns:
            Expression: Output expression with the type ``int32``.
        """
        return self._output_func('CountAll')
    def count_groups(self):
        """Count of all groups in the window.

        Returns:
            Expression: Output expression with the type ``int32``.
        """
        return self._output_func('CountGroups')

    def interval_end(self):
        """Get the end of the current *time-interval* window.

        .. versionadded:: 1.2

        Returns:
            Expression: Output expression with the type ``timestamp``.
        """
        return self._output_func('intervalEnd')

    def interval_start(self):
        """Get the start of the current *time-interval* window.

        .. versionadded:: 1.2

        Returns:
            Expression: Output expression with the type ``timestamp``.
        """
        return self._output_func('intervalStart')

    def pane_timing(self):
        """Get timing of a *time-interval* window pane.

        The timing of a window pane triggering in relation to the enclosing operator's watermark that is used for predicting pane completion.

        * "paneEarly": The system has not yet predicted that it has seen all data which may contribute to a pane's window.
        * "paneOnComplete": The system predicts that it has seen all data which may contribute to a pane's window.
        * "paneLate": The system encountered new data for a pane's window after predicting no more could arrive.

        .. versionadded:: 1.3

        Returns:
            Expression: Output expression with the type ``rstring``.
        """
        return self._output_func('(rstring)paneTiming')

    def max(self, attribute):
        """Maximum value for an input attribute.

        Returns an output expression representing the maximum value
        of the input attribute in the group.

        Example::

            # Get the maximum reading of the last ten tuples grouped by sensor id in the last minute
            schema = 'tuple<int32 id, timestamp ts, float64 max_reading>'

            agg = Aggregate.invoke(s.last(10), schema, group='id')
            agg.max_reading = agg.max('reading')

        Args:
            attribute(str): Attribute name to find the maximum value of.

        Returns:
            Expression: Output expression with the type of the input attribute.
        """
        return self._output_func('Max', attribute)

    def min(self, attribute):
        """Minimum value for an input attribute.

        Returns an output expression representing the minimum value
        of the input attribute in the group.

        Example::

            # Get the minimum reading of the last ten tuples grouped by sensor id
            # updating every input tuple.
            schema = 'tuple<int32 id, timestamp ts, float64 min_reading>'

            agg = Aggregate.invoke(s.last(10), schema, group='id')
            agg.min_reading = agg.min('reading')

        Args:
            attribute(str): Attribute name to find the minimum value of.

        Returns:
            Expression: Output expression with the type of the input attribute.
        """
        return self._output_func('Min', attribute)

    def sum(self, attribute):
        """Sum of values for an input attribute.

        Returns an output expression representing the sum of values
        of the input attribute in the group.

        Example::

            # Get the sum reading of the last ten tuples grouped by sensor id
            # updating every input tuple.
            schema = 'tuple<int32 id, timestamp ts, float64 sum_readings>'

            agg = Aggregate.invoke(s.last(10), schema, group='id')
            agg.sum_readings = agg.sum('reading')

        Args:
            attribute(str): Attribute name to find the sum of.

        Returns:
            Expression: Output expression with the type of the input attribute.
        """
        return self._output_func('Sum', attribute)

    def average(self, attribute):
        """Average of values for an input attribute.

        Returns an output expression representing the average of values
        of the input attribute in the group.

        Example::

            # Get the average reading of the last ten tuples grouped by sensor id
            # updating every input tuple.
            schema = 'tuple<int32 id, timestamp ts, float64 avg_reading>'

            agg = Aggregate.invoke(s.last(10), schema, group='id')
            agg.avg_reading = agg.average('reading')

        Args:
            attribute(str): Attribute name to find the average value of.

        Returns:
            Expression: Output expression with the type of the input attribute.
        """
        return self._output_func('Average', attribute)

    def first(self, attribute):
        """First value for an input attribute.

        Returns an output expression representing the first (earliest) value
        of the input attribute in the group.

        Example::

            # Get the first reading of the last ten tuples grouped by sensor id
            # updating every input tuple.
            schema = 'tuple<int32 id, timestamp ts, float64 first_reading>'

            agg = Aggregate.invoke(s.last(10), schema, group='id')
            agg.first_reading = agg.first('reading')

        Args:
            attribute(str): Attribute name to find the first value of.

        Returns:
            Expression: Output expression with the type of the input attribute.
        """
        return self._output_func('First', attribute)

    def last(self, attribute):
        """Last value for an input attribute.

        Returns an output expression representing the last
        (latest, most recent) value of the input attribute in the group.

        Example::

            # Get the last reading of the last ten tuples grouped by sensor id
            # updating every input tuple.
            schema = 'tuple<int32 id, timestamp ts, float64 last_reading>'

            agg = Aggregate.invoke(s.last(10), schema, group='id')
            agg.last_reading = agg.last('reading')

        Args:
            attribute(str): Attribute name to find the first value of.

        Returns:
            Expression: Output expression with the type of the input attribute.
        """
        return self._output_func('Last', attribute)

    def std(self, attribute, sample=False):
        """Standard deviation of values for an input attribute.

        Returns an output expression representing the standard deviation of values
        of the input attribute in the group.

        Example::

            # Get the standard deviation reading of the last ten tuples grouped by sensor id
            # updating every input tuple.
            schema = 'tuple<int32 id, timestamp ts, float64 sd_reading>'

            agg = Aggregate.invoke(s.last(10), schema, group='id')
            agg.sd_reading = agg.standard_('reading')

        Args:
            attribute(str): Attribute name to find the average value of.
            sample(bool): True to use sample standard deviation otherwise population standard deviation is used.

        Returns:
            Expression: Output expression with the type of the input attribute.
        """
        return self._output_func('SampleStdDev' if sample else 'PopulationStdDev', attribute)


class Filter(Invoke):
    """Removes tuples from a stream by passing along only those tuples that satisfy a user-specified condition.

    Non-matching tuples can be sent to a second optional output.

    The schema transformation is implemented using the ``spl.relational::Filter``
    SPL primitive operator from the SPL Standard toolkit.

    Example with one output stream::

        import streamsx.standard.relational as R
        import streamsx.standard.utility as U

        topo = Topology()
        s = U.sequence(topo, iterations=4)
        matches = R.Filter.matching(s, filter='seq>=2ul')

    Example with matching and non matching streams::

        topo = Topology()
        s = U.sequence(topo, iterations=4)
        matches, non_matches = R.Filter.matching(s, filter='seq<2ul', non_matching=True)

    """
    @staticmethod
    def matching(stream, filter, non_matching=False, name=None):
        """Filters input tuples to one or two output streams

        Args:
            stream: Input stream
            filter(str): Specifies that the condition that determines the tuples to be passed along by the Filter operator
            non_matching(bool): Non-matching tuples are sent to a second optional output stream
            name(str): Invocation name, defaults to a generated name.

        Returns:
             Stream: matching tuples (optional second stream for non matching tuples).
        """
        _op = Filter(stream, non_matching, name=name)
        if filter is not None:
            _op.params['filter'] = _op.expression(filter);
        if non_matching:
            return _op.outputs[0], _op.outputs[1]
        else:
            return _op.outputs[0]

    def __init__(self, stream, non_matching=False, name=None):
        topology = stream.topology
        kind="spl.relational::Filter"
        inputs=stream
        schema = stream.oport.schema
        schemas = [schema,schema] if non_matching else schema
        params = dict()
        super(Filter, self).__init__(topology,kind,inputs,schemas,params,name)


class Functor(Invoke):
    """Transform input tuples into output ones, and optionally filter them.

    If you do not filter an input tuple, any incoming tuple results in a tuple on each output stream

    The schema transformation is implemented using the ``spl.relational::Functor``
    SPL primitive operator from the SPL Standard toolkit.

    Example with schema transformation and two output streams::

        import streamsx.standard.relational as R
        import streamsx.standard.utility as U

        topo = Topology()
        s = U.sequence(topo, iterations=10) # schema is 'tuple<uint64 seq, timestamp ts>'
        fo = R.Functor.map(s, [StreamSchema('tuple<uint64 seq>'),StreamSchema('tuple<timestamp ts>')])
        seq_stream = fo.outputs[0] # schema is 'tuple<uint64 seq>' only
        ts_stream = fo.outputs[1] # schema is 'tuple<timestamp ts>' only

    Example with filter some tuples::

        topo = Topology()
        s = U.sequence(topo, iterations=5)
        fo = R.Functor.map(s, StreamSchema('tuple<uint64 seq>'), filter='seq>=2ul')
        fstream = fo.outputs[0]
        fstream.print()

    """
    @staticmethod
    def map(stream, schema, filter=None, name=None):
        """Map input stream schema to one or more output schemas

        Args:
            stream: Input stream
            schema(str,StreamSchema): Schema of output stream(s).
            filter(str): Specifies the condition that determines which input tuples are to be operated on.
            name(str): Invocation name, defaults to a generated name.

        Returns:
             Functor: Functor invocation.
        """
        _op = Functor(stream, schema, name=name)
        if filter is not None:
            _op.params['filter'] = _op.expression(filter);
        return _op
   
    def __init__(self, stream, schemas, filter=None, name=None):
        topology = stream.topology
        kind="spl.relational::Functor"
        inputs=stream
        params = dict()
        super(Functor, self).__init__(topology,kind,inputs,schemas,params,name)


class Join(Invoke):
    """Correlate tuples from two streams that are based on user-specified match predicates and window configurations.

    The correlation is implemented using the ``spl.relational::Join``
    SPL primitive operator from the SPL Standard toolkit.
    """
    @staticmethod
    def lookup(reference, reference_key, lookup, lookup_key, schema, match=None, name=None):
        """Used to correlate tuples from two streams that are based on user-specified match predicates and window configurations.

        When a tuple is received on an input port, it is inserted into the window corresponding to the input port, which causes the window to trigger.
        As part of the trigger processing, the tuple is compared against all tuples inside the window of the opposing input port.
        If the tuples match, then an output tuple is produced for each match.

        The ``reference_key`` and ``lookup_key`` parameters are used to specify equijoin match predicates, which result in using a hash-based join implementation.

        Args:
            reference(streamsx.topology.topology.Window): Input window
            reference_key(str): Name of the attribute
            lookup(Stream): Input stream
            lookup_key(str): Name of the attribute
            schema(str,StreamSchema): Schema of output stream.
            match(str): Specifies an expression to be used for matching the tuples. The expression might refer to attributes from both input ports. When this parameter is omitted, the default value of true is used.
            name(str): Invocation name, defaults to a generated name.

        Returns:
             Join: Join invocation.
        """
        if not isinstance(reference, streamsx.topology.topology.Window):
            raise TypeError(reference)

        _op = Join(reference, lookup.last(0), schemas=schema, match=match, name=name)
        _op.params['equalityLHS'] = _op.attribute(reference.stream, reference_key)
        _op.params['equalityRHS'] = _op.attribute(lookup, lookup_key)

        if reference._config is not None:
           if 'partitioned' in reference._config:
              if (reference._config['partitioned']):
                 _op.params['partitionByLHS'] = _op.attribute(reference.stream, reference._config['partitionBy'])

        return _op

    def __init__(self, left, right, schemas, match=None, name=None):
        topology = left.topology
        kind="spl.relational::Join"
        inputs = [left, right]
        params = dict()
        if match is not None:
            params['match'] = self.expression(match)
        #if algorithm is not None:
        #    params['algorithm'] = algorithm
        #if defaultTupleLHS is not None:
        #    params['defaultTupleLHS'] = defaultTupleLHS
        #if defaultTupleRHS is not None:
        #    params['defaultTupleRHS'] = defaultTupleRHS
        #if partitionByRHS is not None:
        #    _op.params['partitionByRHS'] = partitionByRHS
        super(Join, self).__init__(topology,kind,inputs,schemas,params,name)





# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2017,2018
"""
Standard utilities for processing streams.
"""

import streamsx.spl.op
from streamsx.topology.schema import StreamSchema
from streamsx.spl.types import float64, uint32, uint64
import streamsx.topology.composite

import streamsx.standard._version
__version__ = streamsx.standard._version.__version__

SEQUENCE_SCHEMA = StreamSchema('tuple<uint64 seq, timestamp ts>')
"""Structured schema containing a sequence identifier and a timestamp.

``'tuple<uint64 seq, timestamp ts>'``
"""

class Sequence(streamsx.topology.composite.Source):
    """A sequence source.

    Creates a structured stream with schema :py:const:`SEQUENCE_SCHEMA` with
    the ``seq`` attribute starting at zero and monotonically increasing and
    ``ts`` attribute set to the time the tuple was generated.

    Args:
        period(float): Period of tuple generation in seconds, if `None` then tuples are generated as fast as possible.
        iterations(int): Number of tuples on the stream, if `None` then the stream is infinite.
        delay(float): Delay in seconds before the first tuple is submitted, if `None` then the tuples are submitted as soon as possible.
        trigger_count(int): Specifies how many tuples are submitted before the operator starts to drain the pipeline of a consistent region and establish a consistent state. 

    Example, create a infinite sequence stream of twenty tuples per second::

        from streamsx.topology.topology import Topology
        import streamsx.standard.utility as U

        topo = Topology()
        seq = topo.source(U.Sequence(period=0.5), name='20Hz')

    Example, start operator of an operator-driven consistent region::

        from streamsx.topology.topology import Topology
        import streamsx.standard.utility as U
        from streamsx.topology.state import ConsistentRegionConfig

        topo = Topology()
        s = topo.source(U.Sequence(iterations=1000, delay=0.1, trigger_count=10))
        s.set_consistent(ConsistentRegionConfig.operator_driven())


    """
    def __init__(self, period:float=None, iterations:int=None, delay:float=None, trigger_count:int=None):
        self.period = period
        self.iterations = iterations
        self.delay = delay
        self.trigger_count = trigger_count

    def populate(self, topology, name, **options):
        return _sequence(topology, self.period, self.iterations, self.delay, self.trigger_count, name)

def _sequence(topology, period=None, iterations=None, delay=None, trigger_count=None, name=None):
    """Create a sequence stream.

    Creates a structured stream with schema :py:const:`SEQUENCE_SCHEMA` with
    the ``seq`` attribute starting at zero and monotonically increasing and
    ``ts`` attribute set to the time the tuple was generated.

    Args:
        period(float): Period of tuple generation in seconds, if `None` then tuples are generated as fast as possible.
        iterations(int): Number of tuples on the stream, if `None` then the stream is infinite.
        delay(float): Delay in seconds before the first tuple is submitted, if `None` then the tuples are submitted as soon as possible.
        trigger_count(int): Specifies how many tuples are submitted before the operator starts to drain the pipeline of a consistent region and establish a consistent state.
        name(str): Name of the stream, if `None` a generated name is used.

    Returns:
        Stream: Structured stream containing an ever increasing ``seq`` attribute.
    """
    if iterations is not None:
        iterations = int(iterations)
    if period is not None:
        period = float(period)
    if name is None:
        name = 'Sequence'
        if iterations is not None:
            name = name + '({:d})'.format(iterations)
        if period is not None:
            name = name + ':period={:.3f}s'.format(period)

    _op = _Beacon(topology, SEQUENCE_SCHEMA, period=period, iterations=iterations, delay=delay, triggerCount=trigger_count, name=name)
    _op.seq = _op.output('IterationCount()')
    _op.ts = _op.output('getTimestamp()')
    return _op.stream

class _Beacon(streamsx.spl.op.Source):
    def __init__(self, topology, schema, period=None, iterations=None, delay=None, triggerCount=None, name=None):
        kind="spl.utility::Beacon"
        inputs=None
        schemas=schema
        params = dict()
        if period is not None:
            params['period'] = float64(period)
        if iterations is not None:
            params['iterations'] = uint32(iterations)
        if delay is not None:
            params['initDelay'] = float64(delay)
        if triggerCount is not None:
            params['triggerCount'] = uint32(triggerCount)
        super(_Beacon, self).__init__(topology,kind,schemas,params,name)


def spray(stream, count, queue=1000, name=None):
    """Spray tuples to a number of streams.
    Each tuple on `stream` is sent to one (and only one)
    of the returned streams.
    The stream for a specific tuple is not defined,
    instead each stream has a dedicated thread and the
    first available thread will take the tuple and
    submit it.

    Each tuple on `stream` is placed on internal queue before it
    is submitted to an output stream. If the queue fills up
    then processing of the input stream is blocked until there
    is space in the queue.

    Example, spray the source tuples to 8 streams::

        from streamsx.topology.topology import Topology
        import streamsx.standard.utility as U

        topo = Topology()
        s = topo.source(U.Sequence())
        outs = []
        for so in U.spray(s, count=8):
            outs.append(so.map(lambda x : (x['seq'], x['ts']), schema=U.SEQUENCE_SCHEMA))
        s = outs[0].union(set(outs))

    Args:
        count(int): Number of output streams the input stream will be sprayed across.
        queue(int): Maximum queue size.
        name(str): Name of the stream, if `None` a generated name is used.

    Returns:
        list(:py:class:`topology_ref:streamsx.topology.topology.Stream`) : List of output streams.
    """
    _op = _ThreadedSplit(stream, count, queue,name=name)
    return _op.outputs


class _ThreadedSplit (streamsx.spl.op.Invoke):
    def __init__(self, stream, count, queue=1000, name=None):
        topology = stream.topology
        kind="spl.utility::ThreadedSplit"
        inputs=stream
        schemas=[stream.oport.schema] * count
        params = dict()
        params['bufferSize'] = uint32(queue)
        super(_ThreadedSplit, self).__init__(topology,kind,inputs,schemas,params,name)

class Throttle(streamsx.topology.composite.Map):
    """Throttle the rate of a stream.

    Args:
         rate(float): Throttled rate of the returned stream in tuples/second.
         precise(bool): Try to make the rate precise at the cost of increased overhead.
         include_punctuations(bool): Specifies whether punctuation is to be included in the rate computation
         period(float): The period to be used for maintaining the wanted rate in seconds. When making rate adjustments, the Throttle operator considers only the last period, going back from the current time. By default, the period is set to 10.0/rate.

    Example throttling a stream ``readings`` to around 10,000 tuples per second::

        import streamsx.standard.utility as U
        readings = readings.map(U.Throttle(rate=10000.0))

    """
    def __init__(self, rate:float, precise:bool=False, include_punctuations:bool=False, period:float=None):
        self.rate = rate
        self.precise = precise
        self.include_punctuations = include_punctuations
        self.period = period

    def populate(self, topology, stream, schema, name, **options):
        _op = _Throttle(stream, self.rate, period=self.period, includePunctuations=self.include_punctuations, precise=self.precise, name=name)
        return _op.stream


class _Throttle (streamsx.spl.op.Map):
    """Stream throttle capability
    """
    def __init__(self, stream, rate, period=None, includePunctuations=None, precise=None, name=None):
        kind="spl.utility::Throttle"
        params = dict()
        params['rate'] = float64(rate)
        if period is not None:
            params['period'] = float64(period)
        if includePunctuations is not None:
            params['includePunctuations'] = includePunctuations
        if precise is not None:
            params['precise'] = precise
        super(_Throttle, self).__init__(kind,stream,params=params,name=name)


def union(inputs, schema, name=None):
    """Union structured streams with disparate schemas.

    Each tuple on any of the streams in `inputs` results in
    a tuple on the returned stream.

    All attributes of the output tuple are set from the input tuple,
    thus the schema of each input must include attributes matching
    (name and type) the output schema.

    The order of attributes in the input schemas need not match
    the output schemas and the input schemas may contain additional
    attributes which will be discarded.

    Example, the output of :py:meth:`~streamsx.standard.utility.union` contains attribute ``c`` only::

        from streamsx.topology.topology import Topology
        import streamsx.standard.utility as U

        topo = Topology()
        ...
        # schema of stream a: 'tuple<int32 a, int32 c>'
        # schema of stream b: 'tuple<int32 c, int32 b>'
        r = U.union([a,b], schema='tuple<int32 c>')


    .. note:: 
        This method differs from :py:meth:`topology_ref:streamsx.topology.topology.Stream.union` in that 
        the schemas of input and output streams can differ, while
        :py:meth:`~streamsx.standard.utility.union` requires matching input and output attributes.

    Args:
        inputs(list[:py:class:`topology_ref:streamsx.topology.topology.Stream`]): Streams to be unioned.
        schema(:py:class:`topology_ref:streamsx.topology.schema.StreamSchema`): Schema of output stream
        name(str): Name of the stream, if `None` a generated name is used.

    Returns:
        :py:class:`topology_ref:streamsx.topology.topology.Stream`: Stream that is a union of `inputs`.

    """
    _op = _Union(inputs, schema, name=name)
    return _op.outputs[0]

class _Union (streamsx.spl.op.Invoke):
    """Union structured streams with disparate schemas.
    """

    def __init__(self, inputs, schema, name=None):
        topology = inputs[0].topology
        kind="spl.utility::Union"
        schemas=schema
        params = None
        super(_Union, self).__init__(topology,kind,inputs,schemas,params,name)


class Deduplicate(streamsx.topology.composite.Map):
    """Deduplicate tuples on a stream.

    If a tuple on `stream` is followed by a duplicate tuple
    within `count` tuples or `period` number of seconds
    then the duplicate is discarded from the returned stream.

    Only one of `count` or `period` can be set.

    Args:
        count(int): Number of tuples.
        period(float): Time period to check for duplicates.
        key(string): Expression used to determine whether a tuple is a duplicate. If this parameter is omitted, the whole tuple is used as the key.
        flush_on_punctuation(bool): Specifies whether punctuation causes the operator to forget all history of remembered tuples. If this parameter is not specified, the default value is False. If the parameter value is True, all remembered keys are erased when punctuation is received.

    Example discarding duplicate tuples wth `a=1` and `a=2`::

        import streamsx.standard.utility as U
        topo = Topology()
        s = topo.source([1,2,1,4,5,2,6,3,7,8,9])
        s = s.map(lambda v : {'a':v}, schema='tuple<int32 a>')
        s = s.map(U.Deduplicate(count=10))

    """
    def __init__(self, count:int=None, period:float=None, key:str=None, flush_on_punctuation:bool=None):
        self.count = count
        self.period = period
        self.key = key
        self.flush_on_punctuation = flush_on_punctuation

    def populate(self, topology, stream, schema, name, **options):
        return _deduplicate(stream, self.count, self.period, self.key, self.flush_on_punctuation, name)

def _deduplicate(stream, count=None, period=None, key=None, flush_on_punctuation=None, name=None):
    if count and period:
        raise ValueError("Cannot set count and period")

    _op = _DeDuplicate(stream, count=count, timeOut=period, key=key, flushOnPunctuation=flush_on_punctuation, name=name)
    return _op.stream

class _DeDuplicate (streamsx.spl.op.Map):
    def __init__(self, stream, timeOut=None, count=None, key=None, flushOnPunctuation=None, name=None):
        kind="spl.utility::DeDuplicate"
        params = dict()
        if timeOut is not None:
            params['timeOut'] = float64(timeOut)
        if count is not None:
            params['count'] = uint64(int(count))
        if key is not None:
            params['key'] = self.expression(key)
        if flushOnPunctuation is not None:
            params['flushOnPunctuation'] = flushOnPunctuation
        #if deltaAttribute is not None:
        #    params['deltaAttribute'] = deltaAttribute
        #if delta is not None:
        #    params['delta'] = delta
        #if resetOnDuplicate is not None:
        #    params['resetOnDuplicate'] = resetOnDuplicate

        super(_DeDuplicate, self).__init__(kind,stream,params=params,name=name)

class Delay(streamsx.topology.composite.Map):
    """Delay tuples on a stream.

    Delays tuples on `stream` maintaining inter-arrival times
    of tuples and punctuation.

    Args:
        delay(float): Seconds to delay each tuple.
        max_delayed(int): Number of items that can be delayed before upstream processing is blocked.

    Example delaying a stream ``readings`` by 1.5 seconds::

        import streamsx.standard.utility as U

        readings = readings.map(U.Delay(delay=1.5))
    """
    def __init__(self, delay:float, max_delayed:int=1000):
        self.delay=delay
        self.max_delayed=max_delayed

    def populate(self, topology, stream, schema, name, **options):
        return _delay(stream, self.delay, self.max_delayed, name)

def _delay(stream, delay, max_delayed=1000, name=None):
    _op = _Delay(stream, delay, max_delayed, name)
    return _op.stream

class _Delay(streamsx.spl.op.Map):
    def __init__(self, stream, delay, max_delayed=1000, name=None):
        topology = stream.topology
        kind="spl.utility::Delay"
        params = dict()
        params['delay'] = float64(delay)
        if max_delayed is not None:
            params['bufferSize'] = uint32(max_delayed)
        super(_Delay, self).__init__(kind,stream,params=params,name=name)

def pair(stream0, stream1, matching=None, buffer_size:int=None, name=None):
    """Pair tuples across two streams.

    This method is used to merge results from performing
    parallel tasks on the same stream, for example perform multiple
    model scoring on the same stream.

    Holds tuples on the two input streams until a matched tuple has been
    received by both input streams. Once matching tuples have received
    the two tuples are submitted to the returned stream with the
    tuple from ``stream0`` followed by the one from ``stream1``.

    Tuples are matched according to the ``matching`` parameter which
    is an attribute name from the input tuple schema,
    typically representing the application key of the tuple.

    If ``matching`` is ``None`` then a match occurs when
    a tuple is received, so that tuples are emitted when a tuple
    has been received by both input streams.
    
    ``stream0`` and ``stream1`` must have the same schema and the resultant
    stream has the same schema.

    These schemas are not supported when ``matching`` is specified.

       * ``CommonSchema.Python``
       * ``CommonSchema.Json``

    This is equivalent to ``merge([stream0, stream1], matching, name)``.

    Example of scoring in parallel::

        import streamsx.standard.utility as U
        
        # Stream of customer information with customer identifier
        # as the id attribute.
        customers = ...
        score_schema = schema.extend(StreamSchema('tuple<float64 score>'))
    
        # Score each tuple on customers in parallel
        cust_churn = s.map(customer_churn_score, schema=score_schema)
        cust_renew = s.map(customer_renew_score, schema=score_schema)
        
        # Pair back as single stream
        # cust_churn_renew stream will contain two tuples for
        # each customer, the churn score followed by the renew score.
        cust_churn_renew = U.pair(cust_churn, cust_renew, matching='id');

    Args:
        stream0(:py:class:`topology_ref:streamsx.topology.topology.Stream`): First input stream.
        stream1(:py:class:`topology_ref:streamsx.topology.topology.Stream`): Second input stream.
        matching(str): Attribute name for matching tuples.
        buffer_size(int): Specifies the size of the internal buffer that is used to queue up tuples from an input port that do not yet have matching tuples from other ports. This parameter is not supported in a consistent region.
        name(str): Name of resultant stream, defaults to a generated name.

    Returns:
        :py:class:`topology_ref:streamsx.topology.topology.Stream`: Paired stream.
    """
    return merge([stream0, stream1], matching, buffer_size, name)

def merge(inputs, matching=None, buffer_size=None, name=None):
    """Merge tuples across two (or more) streams.

    This method is used to merge results from performing
    parallel tasks on the same stream, for example perform multiple
    model scoring on the same stream.

    Holds tuples on the input streams until a matched tuple has been
    received by each input stream. Once matching tuples have received
    for all input streams the tuples are submitted to the returned
    stream in order of the input ports.

    Tuples are matched according to the ``matching`` parameter which
    is an attribute name from the input tuple schema,
    typically representing the application key of the tuple.

    If ``matching`` is ``None`` then a match occurs when
    a tuple is received, so that tuples are emitted when a tuple
    has been received by each input port.
    
    All input streams must have the same schema and the resultant
    stream has the same schema.

    These schemas are not supported when ``matching`` is specified.

       * ``CommonSchema.Python``
       * ``CommonSchema.Json``

    Args:
        inputs(list[:py:class:`topology_ref:streamsx.topology.topology.Stream`]): Input streams to be matched.
        matching(str): Attribute name for matching.
        buffer_size(int): Specifies the size of the internal buffer that is used to queue up tuples from an input port that do not yet have matching tuples from other ports. This parameter is not supported in a consistent region.
        name(str): Name of resultant stream, defaults to a generated name.

    Returns:
        :py:class:`topology_ref:streamsx.topology.topology.Stream`: Merged stream.
    """
    _op = _Pair(inputs, matching, buffer_size=buffer_size, name=name)
    return _op.outputs[0]

class _Pair(streamsx.spl.op.Invoke):
    def __init__(self, inputs, matching=None, buffer_size=None, name=None):
        topology = inputs[0].topology
        kind="spl.utility::Pair"
        schema=inputs[0].oport.schema
        params = dict()
        if buffer_size is not None:
            params['bufferSize'] = uint32(buffer_size)
        super(_Pair, self).__init__(topology,kind,inputs,[schema],params,name)
        if matching is not None:
            for port_idx in range(len(inputs)):
                self.params['partitionBy' + str(port_idx)] = self.attribute(inputs[port_idx], matching)

def gate(stream, control, max_unacked=1, ack_count=1, name=None):
    """Gate tuple flow based upon downstream processing.

    Tuples on `stream` are passed through unmodified to the returned
    stream but the flow is gated by tuples on the `control` stream.

    Up to `max_unacked` tuples flow through the gate before `stream` is
    is blocked. Each tuple arriving on `control` acknowledges `ack_count`
    tuples and thus unblocks `stream` until the number of unacknowledged
    tuples reaches `max_unacked` again.

    The output of some downstream processing is typically used as `control`
    and thus `control` is usually a stream obtained from a :py:class:`topology_ref:streamsx.topology.topology.PendingStream`.

    Example with feedback loop::

        import streamsx.standard.utility as U

        topo = Topology()
        s = topo.source(range(100))
        c = PendingStream(topo)
        g = U.gate(s, c.stream, max_unacked=1)
        g = g.map(lambda _ : time.time())
        r = g.map(U.Delay(delay=1.0))
        c.complete(r)

    Args:
        stream(:py:class:`topology_ref:streamsx.topology.topology.Stream`): Stream to be gated.
        control(:py:class:`topology_ref:streamsx.topology.topology.Stream`): Controlling stream.
        max_unacked(int): Maximum of tuples allowed through the gate without acknowledgement.
        ack_count(int): Count of tuples to acknowledge with each tuple arriving on `control`.
        name(str): Name of resultant stream, defaults to a generated name.
    
    Returns:
        :py:class:`topology_ref:streamsx.topology.topology.Stream`: Gated stream.
    """
    ack_count = uint32(ack_count)
    _op = _Gate([stream,control], maxUnackedTupleCount=max_unacked, numTuplesToAck=ack_count,name=name)
    return _op.outputs[0]

class _Gate(streamsx.spl.op.Invoke):
    def __init__(self, inputs, maxUnackedTupleCount, numTuplesToAck=None, name=None):
        topology = inputs[0].topology
        kind="spl.utility::Gate"
        schema=inputs[0].oport.schema
        params = dict()
        if maxUnackedTupleCount is not None:
            params['maxUnackedTupleCount'] = uint32(maxUnackedTupleCount)
        if numTuplesToAck is not None:
            params['numTuplesToAck'] = numTuplesToAck
        super(_Gate, self).__init__(topology,kind,inputs,[schema],params,name)



from setuptools import setup
import streamsx.standard._version
setup(
  name = 'streamsx.standard',
  packages = ['streamsx.standard'],
  include_package_data=True,
  version = streamsx.standard._version.__version__,
  description = 'Standard toolkit integration for IBM Streams',
  long_description = open('DESC.txt').read(),
  author = 'IBM Streams @ github.com',
  author_email = 'hegermar@de.ibm.com',
  license='Apache License - Version 2.0',
  url = 'https://github.com/IBMStreams/pypi.streamsx.standard',
  keywords = ['streams', 'ibmstreams', 'streaming', 'analytics', 'streaming-analytics'],
  classifiers = [
    'Development Status :: 5 - Production/Stable',
    'License :: OSI Approved :: Apache Software License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
  ],
  install_requires=['streamsx>=1.14.6'],
  
  test_suite='nose.collector',
  tests_require=['nose']
)
